// ***************************************************************
// Copyright (c) 2021 Jittor. All Rights Reserved.
// This file is subject to the terms and conditions defined in
// file 'LICENSE.txt', which is part of this source code package.
// ***************************************************************
#pragma once
#include "var.h"

namespace jittor {

bool check_nan(Var* v);

}